﻿using System;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace TestProject
{
    /// <summary>
    /// 육각형 헬퍼
    /// </summary>
    public static class HexagonHelper
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 육각형 너비 구하기 - GetHexgonWidth(height)

        /// <summary>
        /// 육각형 너비 구하기
        /// </summary>
        /// <param name="height">높이</param>
        /// <returns>육각형 너비</returns>
        public static float GetHexgonWidth(float height)
        {
            return (float)(4 * (height / 2 / Math.Sqrt(3)));
        }

        #endregion
        #region 육각형 인덱스 설정하기 - SetHexgonIndex(x, y, height, row, column)

        /// <summary>
        /// 육각형 인덱스 설정하기
        /// </summary>
        /// <param name="x">X</param>
        /// <param name="y">Y</param>
        /// <param name="height">육각형 높이</param>
        /// <param name="row">행</param>
        /// <param name="column">열</param>
        public static void SetHexgonIndex(float x, float y, float height, out int row, out int column)
        {
            float width = GetHexgonWidth(height);

            column = (int)(x / (width * 0.75f));

            if(column % 2 == 0)
            {
                row = (int)Math.Floor(y / height);
            }
            else
            {
                row = (int)Math.Floor((y - height / 2) / height);
            }

            float testX = column * width * 0.75f;
            float testY = row * height;
 
            if(column % 2 != 0)
            {
                testY += height / 2;
            }

            bool isAbove = false;
            bool isBelow = false;

            float deltaX = x - testX;

            if(deltaX < width / 4)
            {
                float deltaY = y - (testY + height / 2);

                if(deltaX < 0.001)
                {
                    if(deltaY < 0)
                    {
                        isAbove = true;
                    }

                    if(deltaY > 0)
                    {
                        isBelow = true;
                    }
                }
                else if(deltaY < 0)
                {
                    if(-deltaY / deltaX > Math.Sqrt(3))
                    {
                        isAbove = true;
                    }
                }
                else
                {
                    if(deltaY / deltaX > Math.Sqrt(3))
                    {
                        isBelow = true;
                    }
                }
            }

            if(isAbove)
            {
                if(column % 2 == 0)
                {
                    row--;
                }

                column--;
            }
            else if(isBelow)
            {
                if(column % 2 != 0)
                {
                    row++;
                }

                column--;
            }
        }

        #endregion
        #region 포인트 배열 구하기 - GetPointArray(height, row, column, minimumX, minimumY)

        /// <summary>
        /// 포인트 배열 구하기
        /// </summary>
        /// <param name="height">육각형 높이</param>
        /// <param name="row">행</param>
        /// <param name="column">열</param>
        /// <param name="minimumX">최소 X</param>
        /// <param name="minimumY">최소 Y</param>
        /// <returns>포인트 배열</returns>
        public static PointF[] GetPointArray(float height, float row, float column, float minimumX, float minimumY)
        {
            float width = GetHexgonWidth(height);

            float y = height / 2;
            float x = 0;

            y += row * height;

            if(column % 2 != 0)
            {
                y += height / 2;
            }

            x += column * (width * 0.75f);

            return new PointF[]
            {
                new PointF(minimumX + x                , minimumY + y             ),
                new PointF(minimumX + x + width * 0.25f, minimumY + y - height / 2),
                new PointF(minimumX + x + width * 0.75f, minimumY + y - height / 2),
                new PointF(minimumX + x + width        , minimumY + y             ),
                new PointF(minimumX + x + width * 0.75f, minimumY + y + height / 2),
                new PointF(minimumX + x + width * 0.25f, minimumY + y + height / 2)
            };
        }

        #endregion
        #region 육각형 그리드 그리기 - DrawHexgonGrid(graphics, pen, minimumX, maximumX, minimumY, maximumY, height)

        /// <summary>
        /// 육각형 그리드 그리기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="pen">펜</param>
        /// <param name="minimumX">최소 X</param>
        /// <param name="maximumX">최대 X</param>
        /// <param name="minimumY">최소 Y</param>
        /// <param name="maximumY">최대 Y</param>
        /// <param name="height">육각형 높이</param>
        public static void DrawHexgonGrid(Graphics graphics, Pen pen, float minimumX, float maximumX, float minimumY, float maximumY, float height)
        {
            for(int row = 0; ; row++)
            {
                PointF[] pointArray = HexagonHelper.GetPointArray(height, row, 0, minimumX, minimumY);

                if(pointArray[4].Y > maximumY)
                {
                    break;
                }

                for(int column = 0; ; column++)
                {
                    pointArray = HexagonHelper.GetPointArray(height, row, column, minimumX, minimumX);

                    if(pointArray[3].X > maximumX)
                    {
                        break;
                    }

                    if(pointArray[4].Y <= maximumY)
                    {
                        graphics.DrawPolygon(pen, pointArray);
                    }
                }
            }
        }

        #endregion
        #region 다각형 내에서 이미지 그리기 - DrawImageInPolygon(graphcis, pointArray, image)

        /// <summary>
        /// 다각형 내에서 이미지 그리기
        /// </summary>
        /// <param name="graphcis">그래픽스</param>
        /// <param name="pointArray">포인트 배열</param>
        /// <param name="image">이미지</param>
        public static void DrawImageInPolygon(Graphics graphcis, PointF[] pointArray, Image image)
        {
            float minimumX;
            float maximumX;
            float minimumY;
            float maximumY;

            SetPolygonBoundRectangle(pointArray, out minimumX, out maximumX, out minimumY, out maximumY);

            float width   = maximumX - minimumX;
            float height  = maximumY - minimumY;
            float centerX = (minimumX + maximumX) / 2f;
            float centerY = (minimumY + maximumY) / 2f;

            float scaleX = width / image.Width;
            float scaleY = height / image.Height;
            float scale  = Math.Max(scaleX, scaleY);

            float scaledWdth   = image.Width  * scale;
            float scaledHeight = image.Height * scale;
            float radiusX      = scaledWdth   / 2f;
            float radiusY      = scaledHeight / 2f;

            RectangleF sourceRectangle = new RectangleF(0, 0, image.Width, image.Height);

            PointF[] targetPointArray =
            {
                new PointF(centerX - radiusX,  centerY - radiusY),
                new PointF(centerX + radiusX,  centerY - radiusY),
                new PointF(centerX - radiusX,  centerY + radiusY)
            };

            GraphicsPath graphicsPath = new GraphicsPath();

            graphicsPath.AddPolygon(pointArray);

            GraphicsState graphicsState = graphcis.Save();

            graphcis.SetClip(graphicsPath);

            graphcis.DrawImage(image, targetPointArray, sourceRectangle, GraphicsUnit.Pixel);

            graphcis.Restore(graphicsState);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region 다각형 테두리 사각형 설정하기 - SetPolygonBoundRectangle(pointArray, minimumX, maximumX, minimumY, maximumY)

        /// <summary>
        /// 다각형 테두리 사각형 설정하기
        /// </summary>
        /// <param name="pointArray">포인트 배열</param>
        /// <param name="minimumX">최소 X</param>
        /// <param name="maximumX">최대 X</param>
        /// <param name="minimumY">최소 Y</param>
        /// <param name="maximumY">최대 Y</param>
        private static void SetPolygonBoundRectangle(PointF[] pointArray, out float minimumX, out float maximumX, out float minimumY, out float maximumY)
        {
            minimumX = pointArray[0].X;
            maximumX = minimumX;
            minimumY = pointArray[0].Y;
            maximumY = minimumY;

            foreach(PointF point in pointArray)
            {
                if(minimumX > point.X)
                {
                    minimumX = point.X;
                }

                if(maximumX < point.X)
                {
                    maximumX = point.X;
                }

                if(minimumY > point.Y)
                {
                    minimumY = point.Y;
                }

                if(maximumY < point.Y)
                {
                    maximumY = point.Y;
                }
            }
        }

        #endregion
    }
}