﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// 메인 폼
    /// </summary>
    public partial class MainForm : Form
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 육각형 높이
        /// </summary>
        private float hexgonHeight = 200;

        /// <summary>
        /// 테두리 두께
        /// </summary>
        private float borderThickness = 5;
 
        /// <summary>
        /// 육각형 리스트
        /// </summary>
        private List<Hexagon> hexagonList = new List<Hexagon>();

        /// <summary>
        /// 행 수
        /// </summary>
        private int rowCount = 1;
        
        /// <summary>
        /// 열 수
        /// </summary>
        private int columnCount = 1;

        /// <summary>
        /// 비트맵
        /// </summary>
        private Bitmap bitmap = null;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MainForm()

        /// <summary>
        /// 생성자
        /// </summary>
        public MainForm()
        {
            InitializeComponent();

            this.openMenuItem.Click                 += openMenuItem_Click;
            this.saveAsMenuItem.Click               += saveAsMenuItem_Click;
            this.exitMenuItem.Click                 += exitMenuItem_Click;
            this.hexagonHeightTextBox.TextChanged   += hexagonHeightTextBox_TextChanged;
            this.borderThicknessTextBox.TextChanged += hexagonHeightTextBox_TextChanged;
            this.borderColorPictureBox.Click        += borderColorPictureBox_Click;
            this.backgroundColorPictureBox.Click    += backgroundColorPictureBox_Click;
            this.createButton.Click                 += createButton_Click;
            this.pictureBox.MouseMove               += pictureBox_MouseMove;
            this.pictureBox.MouseClick              += pictureBox_MouseClick;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 디렉토리에서 파일 열기 메뉴 항목 클릭시 처리하기 - openMenuItem_Click(sender, e)

        /// <summary>
        /// 디렉토리에서 파일 열기 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void openMenuItem_Click(object sender, EventArgs e)
        {
            if(this.openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                FileInfo fileInfo1 = new FileInfo(this.openFileDialog1.FileName);

                DirectoryInfo directoryInfo = fileInfo1.Directory;

                List<FileInfo> fileInfoList = new List<FileInfo>();

                foreach(FileInfo fileInfo2 in directoryInfo.GetFiles())
                {
                    string fileExtension = fileInfo2.Extension.ToLower().Replace(".", "");

                    if
                    (
                        (fileExtension == "bmp" ) ||
                        (fileExtension == "png" ) ||
                        (fileExtension == "jpg" ) ||
                        (fileExtension == "jpeg") ||
                        (fileExtension == "gif" ) ||
                        (fileExtension == "tiff")
                    )
                    {
                        fileInfoList.Add(fileInfo2);
                    }
                }

                int rowCount    = (int)Math.Sqrt(fileInfoList.Count);
                int columnCount = rowCount;

                if(rowCount * columnCount < fileInfoList.Count)
                {
                    columnCount++;
                }

                if(rowCount * columnCount < fileInfoList.Count)
                {
                    rowCount++;
                }

                this.hexagonList = new List<Hexagon>();

                int index = 0;

                for(int row = 0; row < rowCount; row++)
                {
                    for(int column = 0; column < columnCount; column++)
                    {
                        string name     = fileInfoList[index].Name;
                        string filePath = fileInfoList[index].FullName;

                        Bitmap bitmap = LoadBitmap(filePath);

                        this.hexagonList.Add(new Hexagon(row, column, bitmap, name));

                        index++;

                        if(index >= fileInfoList.Count)
                        {
                            break;
                        }
                    }

                    if(index >= fileInfoList.Count)
                    {
                        break;
                    }
                }

                MakeGrid();
            }
        }

        #endregion
        #region 다른 이름으로 저장 메뉴 항목 클릭시 처리하기 - saveAsMenuItem_Click(sender, e)

        /// <summary>
        /// 다른 이름으로 저장 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void saveAsMenuItem_Click(object sender, EventArgs e)
        {
            if(this.saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                SaveImage(this.bitmap, this.saveFileDialog.FileName);
            }
        }

        #endregion
        #region 종료 메뉴 항목 클릭시 처리하기 - exitMenuItem_Click(sender, e)

        /// <summary>
        /// 종료 메뉴 항목 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void exitMenuItem_Click(object sender, EventArgs e)
        {
            Close();
        }

        #endregion
        #region 육각형 높이 텍스트 박스 텍스트 변경시 처리하기 - hexagonHeightTextBox_TextChanged(sender, e)

        /// <summary>
        /// 육각형 높이 텍스트 박스 텍스트 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void hexagonHeightTextBox_TextChanged(object sender, EventArgs e)
        {
            float height;
            float thickness;

            if(!float.TryParse(this.hexagonHeightTextBox.Text, out height) || height < 10)
            {
                return;
            }

            if(!float.TryParse(this.borderThicknessTextBox.Text, out thickness))
            {
                return;
            }

            this.hexgonHeight    = height;
            this.borderThickness = thickness;

            this.pictureBox.Refresh();
        }

        #endregion
        #region 테두리 색상 픽처 박스 클릭시 처리하기 - borderColorPictureBox_Click(sender, e)

        /// <summary>
        /// 테두리 색상 픽처 박스 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void borderColorPictureBox_Click(object sender, EventArgs e)
        {
            this.colorDialog.Color = this.borderColorPictureBox.BackColor;

            if(this.colorDialog.ShowDialog() == DialogResult.OK)
            {
                this.borderColorPictureBox.BackColor = this.colorDialog.Color;

                this.pictureBox.Refresh();
            }
        }

        #endregion
        #region 배경 색상 픽처 박스 클릭시 처리하기 - backgroundColorPictureBox_Click(sender, e)

        /// <summary>
        /// 배경 색상 픽처 박스 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void backgroundColorPictureBox_Click(object sender, EventArgs e)
        {
            this.colorDialog.Color = this.backgroundColorPictureBox.BackColor;

            if(this.colorDialog.ShowDialog() == DialogResult.OK)
            {
                this.backgroundColorPictureBox.BackColor = this.colorDialog.Color;

                this.pictureBox.Refresh();
            }
        }

        #endregion
        #region 생성 버튼 클릭시 처리하기 - createButton_Click(sender, e)

        /// <summary>
        /// 생성 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void createButton_Click(object sender, EventArgs e)
        {
            MakeGrid();

            this.pictureBox.Visible = true;
        }

        #endregion
        #region 픽처 박스 마우스 이동시 처리하기 - pictureBox_MouseMove(sender, e)

        /// <summary>
        /// 픽처 박스 마우스 이동시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void pictureBox_MouseMove(object sender, MouseEventArgs e)
        {
            int row;
            int column;

            HexagonHelper.SetHexgonIndex(e.X, e.Y, this.hexgonHeight, out row, out column);

            int index = FindHexagon(row, column);

            string indexString = "(" + row.ToString() + ", " + column.ToString() + ")";

            if(index < 0)
            {
                Text = $"Bitmap 클래스 : 스크롤 가능한 육각형 몽타주 비트맵 사용하기 : {indexString}";
            }
            else
            {
                string name = this.hexagonList[index].FilePath;

                int position = name.IndexOf('.');

                Text = $"{indexString} {name.Substring(0, position)}";
            }
        }

        #endregion
        #region 픽처 박스 마우스 클릭시 처리하기 - pictureBox_MouseClick(sender, e)

        /// <summary>
        /// 픽처 박스 마우스 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void pictureBox_MouseClick(object sender, MouseEventArgs e)
        {
            int row;
            int column;

            HexagonHelper.SetHexgonIndex(e.X, e.Y, this.hexgonHeight, out row, out column);

            RemoveHexagon(row, column);

            if(this.openFileDialog2.ShowDialog() == DialogResult.OK)
            {
                Bitmap bitmap = LoadBitmap(openFileDialog2.FileName);

                this.hexagonList.Add(new Hexagon(row, column, bitmap, openFileDialog2.FileName));
            }

            MakeGrid();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region 비트맵 로드하기 - LoadBitmap(filePath)

        /// <summary>
        /// 비트맵 로드하기
        /// </summary>
        /// <param name="filePath">파일 경로</param>
        /// <returns>비트맵</returns>
        private Bitmap LoadBitmap(string filePath)
        {
            using(Bitmap bitmap = new Bitmap(filePath))
            {
                return new Bitmap(bitmap);
            }
        }

        #endregion
        #region 그리드 테두리 사각형 찾기 - FindGridBoundRectangle(maximumX, maximumY)

        /// <summary>
        /// 그리드 테두리 사각형 찾기
        /// </summary>
        /// <param name="maximumX">최대 X</param>
        /// <param name="maximumY">최대 Y</param>
        private void FindGridBoundRectangle(out float maximumX, out float maximumY)
        {
            maximumX = 0;
            maximumY = 0;

            float minimumX = this.borderThickness / 2f;

            PointF[] pointArray;

            pointArray = HexagonHelper.GetPointArray
            (
                this.hexgonHeight,
                this.rowCount - 1,
                this.columnCount - 1,
                minimumX,
                minimumX
            );

            if(maximumX < pointArray[3].X)
            {
                maximumX = pointArray[3].X;
            }

            if(maximumY < pointArray[4].Y)
            {
                maximumY = pointArray[4].Y;
            }

            pointArray = HexagonHelper.GetPointArray
            (
                this.hexgonHeight,
                this.rowCount - 1,
                this.columnCount - 1,
                minimumX,
                minimumX
            );

            if(maximumX < pointArray[3].X)
            {
                maximumX = pointArray[3].X;
            }

            if(maximumY < pointArray[4].Y)
            {
                maximumY = pointArray[4].Y;
            }

            maximumX += minimumX;
            maximumY += minimumX;
        }

        #endregion
        #region 그리드 그리기 - DrawGrid(graphics, maximumX, maximumY)

        /// <summary>
        /// 그리드 그리기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="maximumX">최대 X</param>
        /// <param name="maximumY">최대 Y</param>
        private void DrawGrid(Graphics graphics, int maximumX, int maximumY)
        {
            graphics.Clear(this.backgroundColorPictureBox.BackColor);

            graphics.SmoothingMode = SmoothingMode.AntiAlias;

            float minimumX = this.borderThickness / 2f;

            foreach(Hexagon hexagon in this.hexagonList)
            {
                PointF[] pointArray = HexagonHelper.GetPointArray
                (
                    this.hexgonHeight,
                    hexagon.Row,
                    hexagon.Column,
                    minimumX,
                    minimumX
                );

                if(pointArray[3].X > maximumX)
                {
                    continue;
                }

                if(pointArray[4].Y > maximumY)
                {
                    continue;
                }

                HexagonHelper.DrawImageInPolygon
                (
                    graphics,
                    HexagonHelper.GetPointArray
                    (
                        this.hexgonHeight,
                        hexagon.Row,
                        hexagon.Column,
                        minimumX,
                        minimumX
                    ),
                    hexagon.Bitmap
                );
            }

            using(Pen pen = new Pen(this.borderColorPictureBox.BackColor, this.borderThickness))
            {
                HexagonHelper.DrawHexgonGrid
                (
                    graphics,
                    pen,
                    minimumX,
                    maximumX,
                    minimumX,
                    maximumY,
                    this.hexgonHeight
                );
            }
        }

        #endregion
        #region 그리드 만들기 - MakeGrid()

        /// <summary>
        /// 그리드 만들기
        /// </summary>
        private void MakeGrid()
        {
            if(!int.TryParse(this.rowCountTextBox.Text, out this.rowCount))
            {
                return;
            }

            if(!int.TryParse(this.columnCountTextBox.Text, out this.columnCount))
            {
                return;
            }

            float maximumX;
            float maximumY;

            FindGridBoundRectangle(out maximumX, out maximumY);

            int width  = (int)maximumX + 1;
            int height = (int)maximumY + 1;

            this.bitmap = new Bitmap(width, height);

            using(Graphics graphics = Graphics.FromImage(this.bitmap))
            {
                DrawGrid(graphics, width, height);
            }

            this.pictureBox.Image = this.bitmap;

            this.saveAsMenuItem.Enabled = true;
        }

        #endregion
        #region 이미지 저장하기 - SaveImage(image, filePath)

        /// <summary>
        /// 이미지 저장하기
        /// </summary>
        /// <param name="image">이미지</param>
        /// <param name="filePath">파일 경로</param>
        private void SaveImage(Image image, string filePath)
        {
            string fileExtension = Path.GetExtension(filePath);

            switch(fileExtension.ToLower())
            {
                case ".bmp"  : image.Save(filePath, ImageFormat.Bmp ); break;
                case ".exif" : image.Save(filePath, ImageFormat.Exif); break;
                case ".gif"  : image.Save(filePath, ImageFormat.Gif ); break;
                case ".jpg"  :
                case ".jpeg" : image.Save(filePath, ImageFormat.Jpeg); break;
                case ".png"  : image.Save(filePath, ImageFormat.Png ); break;
                case ".tif"  :
                case ".tiff" : image.Save(filePath, ImageFormat.Tiff); break;
                default      : throw new NotSupportedException($"알 수 없는 파일 확장자 : {fileExtension}");
            }
        }

        #endregion
        #region 육각형 찾기 - FindHexagon(row, column)

        /// <summary>
        /// 육각형 찾기
        /// </summary>
        /// <param name="row">행</param>
        /// <param name="column">열</param>
        /// <returns>인덱스</returns>
        private int FindHexagon(int row, int column)
        {
            for (int i = this.hexagonList.Count - 1; i >= 0; i--)
            {
                if((this.hexagonList[i].Row == row) && (this.hexagonList[i].Column == column))
                {
                        return i;
                }
            }

            return -1;
        }

        #endregion
        #region 육각형 제거하기 - RemoveHexagon(row, column)

        /// <summary>
        /// 육각형 제거하기
        /// </summary>
        /// <param name="row">행</param>
        /// <param name="column">열</param>
        private void RemoveHexagon(int row, int column)
        {
            int index = FindHexagon(row,column);

            if(index >= 0)
            {
                this.hexagonList.RemoveAt(index);
            }
        }

        #endregion
    }
}